from pyH2A.Utilities.input_modification import insert, process_table

class Production_Scaling_Plugin:
	'''Calculation of plant output and potential scaling.

	Parameters
	----------
	Technical Operating Parameters and Specifications > Plant Design Capacity (kg of H2/day) > Value : float
		Plant design capacity in kg of H2/day, ``process_table()`` is used.
	Technical Operating Parameters and Specifications > Operating Capacity Factor (%) > Value : float
		Operating capacity factor in %, ``process_table()`` is used.
	Technical Operating Parameters and Specifications > Maximum Output at Gate > Value : float, optional
		Maximum output at gate in (kg of H2)/day, ``process_table()`` is used. If this parameter is
		not specified it defaults to `Plant Design Capacity (kg of H2/day)`.
	Technical Operating Parameters and Specifications > New Plant Design Capacity (kg of H2/day) > Value : float, optional
		New plant design capacity in kg of H2/day to calculate scaling, which overwrites possible Scaling Ratio,
		``process_table()`` is used.
	Technical Operating Parameters and Specifications > Scaling Ratio > Value : float, optional
		Scaling ratio which is multiplied by current plant design capacity to obtain scaled plant size,
		``process_table`` is used.
	Technical Operating Parameters and Specifications > Capital Scaling Exponent > Value : float, optional
		Exponent to calculate capital scaling factor, ``process_table()`` is used. Defaults to 0.78.
	Technical Operating Parameters and Specifications > Labor Scaling Exponent > Value : float, optional
		Exponent to calculcate labor scaling factor, ``process_table()`` is used. Defaults to 0.25.

	Returns
	-------
	Technical Operating Parameters and Specifications > Design Output per Day > Value : float
		Design output in (kg of H2)/day.
	Technical Operating Parameters and Specifications > Max Gate Output per Day > Value : float
		Maximum gate ouput in (kg of H2)/day.
	Technical Operating Parameters and Specifications > Output per Year > Value : float
		Yearly output taking operating capacity factor into account, in (kg of H2)/year.
	Technical Operating Parameters and Specifications > Output per Year at Gate > Value	: float
		Actual yearly output at gate, in (kg of H2)/year.
	Technical Operating Parameters and Specifications > Scaling Ratio > Value : float or None
		Returned if New Plant Design Capacity was specified.
	Scaling > Capital Scaling Factor > Value : float or None
		Returned if scaling is active (`Scaling Ratio` or `New Plant Design Capacity (kg of H2/day)` specified).
	Scaling > Labor Scaling Factor > Value : float or None
		Returned if scaling is active (`Scaling Ratio` or `New Plant Design Capacity (kg of H2/day)` specified).

	Notes
	-----
	To scale capital or labor costs, a path to `Scaling > Capital Scaling Factor > Value`
	or `Scaling > Labor Scaling Factor > Value` has to specified for the respective table entry.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')

		self.dictionary = dcf.inp['Technical Operating Parameters and Specifications']

		self.calculate_scaling(dcf, print_info)
		self.calculate_output(dcf)

		insert(dcf, 'Technical Operating Parameters and Specifications', 'Design Output per Day', 'Value', 
				self.design_output_per_day, __name__, print_info = print_info)
		insert(dcf, 'Technical Operating Parameters and Specifications', 'Max Gate Output per Day', 'Value', 
				self.max_gate_output_per_day, __name__, print_info = print_info)

		insert(dcf, 'Technical Operating Parameters and Specifications', 'Output per Year', 'Value', 
				self.output_per_year, __name__, print_info = print_info)
		insert(dcf, 'Technical Operating Parameters and Specifications', 'Output per Year at Gate', 'Value', 
				self.output_per_year_at_gate, __name__, print_info = print_info)

	def calculate_scaling(self, dcf, print_info):
		'''Calculation of scaling if scaling is requested (either `New Plant Design Capacity (kg of H2/day)` or
		`Scaling Ratio` was provided). Otherwise returns regular design output and output at gate per day in (kg H2).
		'''

		if 'Maximum Output at Gate' not in self.dictionary:
			maximum_output_at_gate = self.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			insert(dcf, 'Technical Operating Parameters and Specifications', 'Maximum Output at Gate', 'Value', 
					maximum_output_at_gate, __name__, print_info = print_info)
	
		if 'New Plant Design Capacity (kg of H2/day)' in self.dictionary:
			scaling_ratio = self.dictionary['New Plant Design Capacity (kg of H2/day)']['Value'] / self.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			insert(dcf, 'Technical Operating Parameters and Specifications', 'Scaling Ratio', 'Value', 
					scaling_ratio, __name__, print_info = print_info)

		if 'Scaling Ratio' in self.dictionary:
			self.design_output_per_day = self.dictionary['Plant Design Capacity (kg of H2/day)']['Value'] * self.dictionary['Scaling Ratio']['Value']
			self.max_gate_output_per_day = self.dictionary['Maximum Output at Gate']['Value'] * self.dictionary['Scaling Ratio']['Value']

			if 'Capital Scaling Exponent' in self.dictionary:
				self.capital_scaling_factor = self.dictionary['Scaling Ratio']['Value'] ** self.dictionary['Capital Scaling Exponent']['Value']
			else:
				self.capital_scaling_factor = self.dictionary['Scaling Ratio']['Value'] ** 0.78

			if 'Labor Scaling Exponent' in self.dictionary:
				self.labor_scaling_factor = self.dictionary['Scaling Ratio']['Value'] ** self.dictionary['Labor Scaling Exponent']['Value']
			else:
				self.labor_scaling_factor = self.dictionary['Scaling Ratio']['Value'] ** 0.25

			insert(dcf, 'Scaling', 'Capital Scaling Factor', 'Value', self.capital_scaling_factor, __name__, print_info = print_info)
			insert(dcf, 'Scaling', 'Labor Scaling Factor', 'Value', self.labor_scaling_factor, __name__, print_info = print_info)

		else:
			self.design_output_per_day = self.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			self.max_gate_output_per_day = self.dictionary['Maximum Output at Gate']['Value']

	def calculate_output(self, dcf):
		'''Calculation of yearly output in kg and yearly output at gate in kg.
		'''

		self.output_per_year = self.design_output_per_day * 365. * self.dictionary['Operating Capacity Factor (%)']['Value']
		self.output_per_year_at_gate = self.max_gate_output_per_day * 365. * self.dictionary['Operating Capacity Factor (%)']['Value']
		


import click
from pyH2A.run_pyH2A import command_line_pyH2A
from pyH2A.Utilities.plugin_input_output_processing import Generate_Template_Input_File

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
	'''Command line interface wrapper function.
	'''
	pass

@cli.command()
@click.option('-i', '--input_file', type=str, help='Path to input file.', required = True)
@click.option('-o', '--output_dir', type=str, help='Path to output directory.', required = True)
def run(input_file, output_dir):
	'''Run pyH2A analysis.
	'''
	output = command_line_pyH2A(input_file, output_dir)

@cli.command()
@click.option('-i', '--input_file', type=str, help='Path to input file.', required = True)
@click.option('-o', '--output_file', type=str, help='Path to output file.', required = True)
@click.option('--origin/--no-origin', default = False, help='Include information on which plugin/module requests input.')
@click.option('--comments/--no-comments', default = False, help='Include comments/documentation for each requested input.')
def generate(input_file, output_file, origin, comments):
	'''Generate input file template from minimal input file.
	'''
	Generate_Template_Input_File(input_file, output_file, origin = origin, comment = comments)







import numpy as np
from pyH2A.Utilities.input_modification import insert, process_table

class Solar_Concentrator_Plugin:
	'''Simulation of solar concentration (used in combination with PEC cells).

	Parameters
	----------
	Solar Concentrator > Concentration Factor > Value : float
		Concentration factor of solar concentration, value > 1.
	Solar Concentrator > Cost ($/m2) > Value : float
		Cost of solar concentrator in $/m2.
	PEC Cells > Number > Value : float
		Number of PEC cells required for design H2 production capacity.
	Land Area Requirement > South Spacing (m) > Value : float
		South spacing of solar concentrators in m.
	Land Area Requirement > East/West Spacing (m) > Value : float
		East/West Spacing (m) of solar concentrators in m.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Total solar collection area in m2.

	Returns
	-------
	Non-Depreciable Capital Costs > Land required (m2) > Value : float
		Total land requirement in m2.
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land requirement in acres.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Total solar collection area in m2.
	Direct Capital Costs - Solar Concentrator > Solar Concentrator Cost ($) > Value : float
		Total cost of all solar concentrators.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Solar Concentrator', 'Value')
		process_table(dcf.inp, 'PEC Cells', 'Value')
		process_table(dcf.inp, 'Land Area Requirement', 'Value')
		process_table(dcf.inp, 'Non-Depreciable Capital Costs', 'Value')

		self.land_area(dcf)
		self.calculate_cost(dcf)

		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (m2)', 'Value',
				self.total_land_area_m2, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (acres)', 'Value', 
				self.total_land_area_acres, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Solar Collection Area (m2)', 'Value', 
				self.total_solar_collection_area_m2, __name__, print_info = print_info)
		
		insert(dcf, 'Direct Capital Costs - Solar Concentrator', 'Solar Concentrator Cost ($)', 'Value', 
				self.concentrator_cost, __name__, print_info = print_info)

	def land_area(self, dcf):
		'''Calculation of solar collection area by multiplying concentration factor by supplied
		(unconcentrated) solar collection area. Calculation of total land area requirement based
		on number of PEC cells and spacing of solar concentrators.
		'''

		land = dcf.inp['Land Area Requirement']

		self.total_solar_collection_area_m2 = dcf.inp['Solar Concentrator']['Concentration Factor']['Value'] * dcf.inp['Non-Depreciable Capital Costs']['Solar Collection Area (m2)']['Value']

		area_per_element_m2 = self.total_solar_collection_area_m2 / dcf.inp['PEC Cells']['Number']['Value']
		side_length_m = np.sqrt(area_per_element_m2)

		x_length_m = side_length_m + land['East/West Spacing (m)']['Value']/2.
		y_length_m = side_length_m + land['South Spacing (m)']['Value']/2.

		spaced_area_per_element_m2 = x_length_m * y_length_m

		self.total_land_area_m2 = spaced_area_per_element_m2 * dcf.inp['PEC Cells']['Number']['Value']
		#self.total_land_area_m2 = self.total_solar_collection_area_m2 + land['South Spacing (m)']['Value'] * land['East/West Spacing (m)']['Value'] * dcf.inp['PEC Cells']['Number']['Value']
		self.total_land_area_acres = self.total_land_area_m2 * 0.000247105

	def calculate_cost(self, dcf):
		'''Calculation of solar concentrator cost based on cost per m2 and total solar collection area.
		'''

		self.concentrator_cost = dcf.inp['Solar Concentrator']['Cost ($/m2)']['Value'] * self.total_solar_collection_area_m2



from pathlib import Path

from pyH2A.Utilities.input_modification import insert, convert_input_to_dictionary, check_for_meta_module, import_plugin, merge, parse_parameter
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow

def is_parameter_or_output(line, spaces_for_tab = 4, spaces_cutoff = 5):
	'''Detection of parameters and output values in line based on presence of more than `spaces_cuttoff`
	spaces (tabs are converted to four spaces).
	'''

	line = line.replace('\t', ' ' * spaces_for_tab) # replace tabs with given number of spaces
	return line[:spaces_cutoff] != ' ' * spaces_cutoff

def process_single_line(line, output_dict, origin, variable_string, **kwargs):
	'''Process single line to extract parameter/output information and comments'''

	if is_parameter_or_output(line, **kwargs): # is a line containing parameter/output information
		complete_string = line.split(':')
		variable_string = complete_string[0].strip(', \t')
		variable_type = complete_string[1].strip(', \t')

		output_dict[variable_string] = {'Type': variable_type, 'Origin': origin}

		return variable_string

	else:  # is a line containing comment for parameter/output
		comment = line.strip(' \t')
		bottom_key = parse_parameter(variable_string)[-1]
		comment_string = f'Comment {bottom_key}'

		if comment_string in output_dict[variable_string]:
			output_dict[variable_string][comment_string] += ' ' + comment
		else:
			output_dict[variable_string][comment_string] = comment

		return variable_string

def extract_input_output_from_docstring(target, **kwargs):
	'''Convert docstring to structured dictionary.'''

	doc_string = target.__doc__.split('\n')

	parameters_dict = {}
	output_dict = {}

	parameters = False
	output = False
	variable_string = None

	for counter, line in enumerate(doc_string):

		if '---' in line and 'Parameters' in doc_string[counter-1]:
			parameters = True

		if '---' in line and 'Returns' in doc_string[counter-1]:
			parameters = False
			output = True
	
		if line.strip(' \t') == '':
			parameters = False
			output = False

		if parameters and '---' not in line:
			variable_string = process_single_line(line, parameters_dict, 
												  target.__name__, variable_string, 
												  **kwargs)

		if output and '---' not in line:
			variable_string = process_single_line(line, output_dict, 
												  target.__name__, variable_string, 
												  **kwargs)

	plugin_dict = {'Parameters': parameters_dict, 'Output': output_dict}

	return plugin_dict

def convert_inp_to_requirements(dictionary, path = None):
	'''Convert inp dictionary structure to requirements dictionary structure.'''
	
	output = {}

	for top_key, top_item in dictionary.items():
		for middle_key, middle_item in top_item.items():
			for bottom_key, bottom_item in middle_item.items():

				path = f'{top_key} > {middle_key} > {bottom_key}'
				output[path] = {'Entry': bottom_item, 'Origin': 'Input'}

	return output

class Generate_Template_Input_File:
	'''Generate input file template from a minimal input file.

	Parameters
	----------
	input_file_stub : str
		Path to input file containing workflow and analysis specifications.
	output_file : str
		Path to file where input template is to be written.
	origin : bool, optional
		Include origin of each requested input parameter in input template
		file ("requested by" information).
	comment : bool, optional
		Include comments for each requested input parameter (additional
		information on parameter).

	Returns
	-------
	Template : object
		Template object which contains information on requirements and
		output. Input template is written to specified output file.
	'''

	def __init__(self, input_file_stub, output_file, 
				 origin = False, comment = False):
		if isinstance(input_file_stub, str):
			self.inp_stub = convert_input_to_dictionary(input_file_stub)
		else:
			self.inp_stub = input_file_stub

		self.inp = {}

		post_workflow_position = self.get_post_workflow_position()

		pre_workflow = {'Description': 'Functions executed before workflow.',
						'Position': -1,
						'Type': 'function'}
		post_workflow = {'Description': 'Function executed after workflow.',
						 'Position': post_workflow_position, 
						 'Type': 'function'}

		self.inp_stub['Workflow']['pre_workflow'] = pre_workflow
		self.inp_stub['Workflow']['post_workflow'] = post_workflow
		self.get_analysis_modules(post_workflow_position)

		self.sorted_keys = sorted(self.inp_stub['Workflow'], 
								  key = lambda x: self.inp_stub['Workflow'][x]['Position'])

		self.requirements = {}
		self.output = {}

		self.provided_inp = convert_inp_to_requirements(self.inp_stub)
		self.generate_requirements()
		self.convert_requirements_to_inp(insert_origin = origin, insert_comment = comment)

		self.inp = merge(self.inp, 
						 convert_input_to_dictionary(input_file_stub, 
						 							 merge_default = False))

		template_file = Template_File(self.inp)
		template_file.write_template_file(output_file)

	def get_post_workflow_position(self):

		max_position = max(item['Position'] for item in self.inp_stub['Workflow'].values()) 

		return max_position + 1

	def get_analysis_modules(self, post_workflow_position):
		'''Get analysis modules from input stub.
		'''

		position = post_workflow_position

		for key in self.inp_stub:
			if check_for_meta_module(key):
				position += 1
				module = {'Description': 'meta module',
						  'Position': position,
						  'Type': 'analysis'}

				self.inp_stub['Workflow'][key] = module

	def generate_requirements(self):
		'''Generate dictionary with input requirements.
		'''

		output = self.provided_inp
		requirements = {}

		for key in self.sorted_keys:
			data = self.get_docstring_data(key, self.inp_stub['Workflow'][key]['Type'])

			needed_parameters = self.check_parameters(data['Parameters'], output)
			requirements = merge(requirements, needed_parameters)

			output = merge(output, data['Output'])

		self.requirements = requirements

	def get_docstring_data(self, target_name, target_type):
		'''Get parameter requirements and outputs from docstrings.
		'''

		if target_type == 'function':
			target = getattr(Discounted_Cash_Flow, target_name)
			data = extract_input_output_from_docstring(target, spaces_cutoff = 9)

		else:
			if target_type == 'plugin':
				plugin_module = True
			else:
				plugin_module = False

			target = import_plugin(target_name, plugin_module)
			data = extract_input_output_from_docstring(target)

		return data

	def check_parameters(self, data, output):
		'''Check if needed parameter is in output.
		'''

		requirements = {}

		for key, item in data.items():
			if key not in output:
				requirements[key] = item
			else:
				item['Fullfilled by'] = output[key]['Origin']

				try:
					if output[key]['Type'] != item['Type']:
						print('Warning: type of output and requirement differs for: `{0}`. `{1}` required, `{2}` provided'.format(key, item['Type'], output[key]['Type']))
				except KeyError:
					pass

		return requirements

	def convert_requirements_to_inp(self, insert_origin = False, insert_comment = False):
		'''Convert dictionary of requirements to formatted self.inp
		'''

		for key, item in self.requirements.items():
			path = parse_parameter(key)
			path = ['[...]' if x == '' else x for x in path] # replacing '>>' with '> [...] >'

			insert(self, *path, item['Type'], None, print_info = False, 
				  add_processed = False, insert_path = False)

			if insert_origin:
				insert(self, *path[:-1], 'Requested by', item['Origin'], None, 
					   print_info = False, add_processed = False, insert_path = False)

			if insert_comment:
				for key in item:
					if 'Comment' in key:
						insert(self, *path[:-1], key, item[key], None, 
							   print_info = False, add_processed = False, insert_path = False)
		
class Template_File:
	'''Generate input file from inp dictionary.

	Parameters
	----------
	inp : dict
		Dictionary containing information on requested input (generated by
		`Generate_Template_Input_File`)

	Returns
	-------
	Template_File : object
		Object containing formatted string for output file.
	'''

	def __init__(self, inp):
		self.inp = inp
		self.convert_inp_to_string()

	def convert_inp_to_string(self):
		'''Convert inp to string.'''

		output = ''

		for top_key, top_item in self.inp.items():
			output += f'# {top_key}'

			columns = list(top_item)
			column_names = self.get_column_names(top_item)

			output += '\n\n' + self.convert_column_names_to_string(column_names)
			output += '\n' + '--- | ' * (len(column_names) - 1) + '---'
			output += self.get_row_entries(columns, column_names, top_item)

		self.output = output

	def get_column_names(self, dictionary):
		'''Get names of table columns.'''

		column_names_array = [['Parameter']]

		for key, item in dictionary.items():
			column_names_array.append(list(item))

		column_names_flat_list = [item for sublist in column_names_array for item in sublist]
		column_names = list(dict.fromkeys(column_names_flat_list))

		return column_names

	def convert_column_names_to_string(self, names):
		'''Convert list of names to markdown style table string.'''

		string = ''

		for name in names[:-1]:
			string += f'{name} | '

		string += names[-1]

		return string

	def get_row_entries(self, columns, column_names, dictionary):
		'''Get entries for each row of table.'''

		string = ''

		for column in columns:
			string += f'\n{column} | '
			string += self.get_single_row(column_names, column, dictionary)

		string += '\n\n'

		return string

	def get_single_row(self, column_names, column, dictionary):
		'''Get entries for a single row.'''

		string = ''

		for counter, name in enumerate(column_names[1:]):
			try:
				entry = dictionary[column][name]
			except KeyError:
				entry = 'n/a'

			string += str(entry) 

			if counter < len(column_names[1:]) - 1:
				string += ' | '

		return string

	def write_template_file(self, file_name):
		'''Write output string to file.'''

		with open(Path(file_name), 'w') as text_file:
			text_file.write(self.output)


		


import numpy as np
import math as math

def find_nearest(array, values):
    '''Find value(s) in `array` that are nearest to `values`.

    Parameters
    ----------
    array: ndarray
        Array to be searched. If `array` has more than one dimension, only
        the first column is used.
    values : float, ndarray or list
        Single float, ndarray or list with values for which the nearest entries
        in `array` should be found.

    Returns
    -------
    hits : list
        List of indices of closest values in `array`.
    '''
    
    if array.ndim != 1:
        array_1d = array[:,0]
    else:
        array_1d = array

    values = np.atleast_1d(values)
    hits = []

    for i in range(len(values)):

        idx = np.searchsorted(array_1d, values[i], side= "left")
        if idx > 0 and (idx == len(array_1d) or math.fabs(values[i] - array_1d[idx-1]) < math.fabs(values[i] - array_1d[idx])):
            hits.append(idx-1)
        else:
            hits.append(idx)

    return hits


from pyH2A.Utilities.input_modification import insert, process_input, sum_all_tables, process_table
import pyH2A.Utilities.find_nearest as fn
import numpy as np

class Replacement_Plugin:
	'''Calculating yearly overall replacement costs based on one-time replacement costs and frequency.

	Parameters
	----------
	Planned Replacement > [...] > Frequency (years) : float
		Replacement frequency of [...] in years. 
		Iteration over all entries in `Planned Replacement` table. No path key
		available.
	Planned Replacement > [...] > Cost ($) : float
		One-time replacement cost of [...].
		Iteration over all entries in `Planned Replacement` table. Path key
		is 'Path'.
	[...] Unplanned Replacement [...] >> Value : float
		``sum_all_tables()`` is used.

	Returns
	-------
	[...] Unplanned Replacement [...] > Summed Total > Value : float
		Summed total for each individual table in "Unplanned Replacement" group.
	Replacement > Total > Value : ndarray
		Total inflated replacement costs (sum of `Planned Replacement` entries and
		unplanned replacement costs).
	'''
	def __init__(self, dcf, print_info):
		# process_table(dcf.inp, 'Planned Replacement', ['Cost ($)', 'Frequency (years)'], 
		# 				path_key = ['Path', 'Frequency Path'])

		self.initialize_yearly_costs(dcf)
		self.calculate_planned_replacement(dcf)
		self.unplanned_replacement(dcf, print_info)

		yearly_inflated = self.yearly * dcf.inflation_correction * dcf.inflation_factor


		insert(dcf, 'Replacement', 'Total', 'Value', yearly_inflated, __name__, print_info = print_info)

	def initialize_yearly_costs(self, dcf):
		'''Initializes ndarray filled with zeros with same length as dcf.inflation_factor.
		'''

		self.yearly = np.zeros(len(dcf.inflation_factor))

	def calculate_planned_replacement(self, dcf):
		'''Calculation of yearly replacement costs by iterating over all entries of 
		`Planned Replacement`.
		'''

		for key in dcf.inp['Planned Replacement']:
			planned_replacement = Planned_Replacement(dcf.inp['Planned Replacement'][key], key, dcf)
			self.yearly[planned_replacement.years_idx] += planned_replacement.cost

	def unplanned_replacement(self, dcf, print_info):
		'''Calculating unplanned replacement costs by appling ``sum_all_tables()`` to 
		"Unplanned Replacement" group.
		'''

		self.unplanned = sum_all_tables(dcf.inp, 'Unplanned Replacement', 'Value', 
										insert_total = True, class_object = dcf, 
										print_info = print_info)
		self.yearly += self.unplanned

class Planned_Replacement:
	'''
	Individual planned replacement objects.

	Methods
	-------
	calculate_yearly_cost:
		Calculation of yearly costs from one-time cost and replacement frequency.
	'''

	def __init__(self, dictionary, key, dcf):
		self.calculate_yearly_cost(dictionary, key, dcf)
		
	def calculate_yearly_cost(self, dictionary, key, dcf):
		'''Calculation of yearly replacement costs.

		Replacement costs are billed annually, replacements which are performed at a non-integer rate 
		are corrected using non_integer_correction.
		'''

		replacement_frequency = int(np.ceil(dictionary['Frequency (years)']))
		non_integer_correction = replacement_frequency / dictionary['Frequency (years)']

		raw_replacement_cost = process_input(dcf.inp, 'Planned Replacement', key, 'Cost ($)')
		initial_replacement_year_idx = fn.find_nearest(dcf.plant_years, replacement_frequency)[0]

		self.cost = raw_replacement_cost * non_integer_correction * dcf.combined_inflator
		self.years = dcf.plant_years[initial_replacement_year_idx:][0::replacement_frequency]
		self.years_idx = fn.find_nearest(dcf.plant_years, self.years)


from scipy import constants as con

def nm(value):
	'''Converts nm to J'''
	return ((con.h*con.c)/(value/1e9))

def eV(value):
	'''Converts eV to J'''
	return value*1.602176e-19

def kcalmol(value):
	'''Converts kcal/mol to J'''
	return (value * 4186.798188)/con.Avogadro

def Jmol(value):
	'''Converts J/mol to J'''
	return value / con.Avogadro

def kWh(value):
	'''Converts kWh to J'''
	return value * 3.6e6

def J(value):
	'''Converts J to J'''
	return value

def kJmol(value):
	'''Converts kJ/mol to J'''
	return (1e3 * value) / con.Avogadro

class Energy:
	'''Energy class to convert between different energy units.

	Parameters
	----------
	value : float
		Input energy value.
	unit : function
		Unit name which corresponds to the one of the functions defined
		outside of the class. This function is used to convert
		the input energy value to Joule.

	Notes
	-----
	Input value in either nm, eV, kcal/mol, J/mol, kJ/mol, kWh or J.
	Available units: J, eV, nm, kcal/mol, J/mol, kWh, kJ/mol.
	Once an Energy class object has been generated, the energy 
	value in the desired unit can be retrieved using the appropriate class
	attribute.
	'''

	def __init__(self, value, unit):
		self.unit = unit.__name__
		self.value = value

		self.J = unit(self.value)
		self.eV = self.convert_J_to_eV()
		self.nm = self.convert_J_to_nm()
		self.kcalmol = self.convert_J_to_kcalmol()
		self.Jmol = self.convert_J_to_Jmol()
		self.kWh = self.convert_J_to_kWh()
		self.kJmol = self.convert_J_to_kJmol()

	def convert_J_to_eV(self):
		'''Convert J to eV'''
		return self.J/1.602176e-19

	def convert_J_to_nm(self):
		'''Convert J to nm'''
		return ((con.h*con.c)/self.J)*1e9

	def convert_J_to_kcalmol(self):
		'''Convert J to kcal/mol'''
		return (self.J * con.Avogadro)/4186.798188

	def convert_J_to_Jmol(self):
		'''Convert J to J/mol'''
		return self.J * con.Avogadro

	def convert_J_to_kWh(self):
		'''Convert J to kWh'''
		return self.J/3.6e6

	def convert_J_to_kJmol(self):
		'''Convert to J to kJ/mol'''
		return (self.J * con.Avogadro) / 1e3


import types
import numpy as np
import matplotlib.pyplot as plt
from pyH2A.Utilities.output_utilities import make_bold, millify, bottom_offset, Figure_Lean
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow

class Cost_Contributions_Analysis:
	'''Reading cost contributions from DCF output and plotting it.

	Notes
	-----
	`Cost_Contributions_Analysis` does not require any input, it is sufficient
	to include "# Cost_Contributions_Analysis" in input file.
	Contributions for plotting have to provided as a dictionary with three components:
	`Data` containing the name and value for each contribution.
	`Total` containing the total value
	`Table Group`, containing the name of overarching group of values, which are shown. 
	Dictionaries with this structure can be automatically generated by ``sum_all_tables`` 
	function.
	If the name of a contribution contains a '-', the string will be split and only the part after the
	'-' will be displayed.
	'''

	def __init__(self, input_file):
		self.base_case = Discounted_Cash_Flow(input_file, print_info = False)

	def cost_breakdown_plot(self, ax = None, figure_lean = True,
							plugin = None, plugin_property = None,
							label_offset = 5.5, x_label_string = 'Cost / USD',
							x_label_string_H2 = r'Levelized cost / USD per kg $H_{2}$',
							plot_kwargs = {}, **kwargs):
		'''Plotting cost breakdown plot.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		plugin : str or None, optional
			Selection of plugin from which contributions data is to be read. If None,
			defaults to overall H2 cost contributions.
		plugin_property: str or None, optional
			Specific property of plugin from which contributions data is to be read.
		label_offset : float, optional
			Offset for contributions labels.
		x_label_string : str, optional
			String for x axis label.
		x_label_string_H2 : str, optional
			String for x axis label when overall H2 cost breakdown is plotted
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.

		Notes
		-----
		Names of contributions are split at '-' symbols and only the last
		part is displayed. For example: `Direct Capital Costs - Installation`
		is displayed only as `Installation`.
		'''

		kwargs = {**{'left': 0.4, 'top': 0.96, 
					 'name': 'Cost_Breakdown_Plot'}, 
				  **kwargs, **plot_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		if plugin is None:
			self.contributions = self.base_case.contributions
			self.data = self.base_case.contributions['Data']
		else:
			selected_plugin = self.base_case.plugs[plugin]
			self.contributions = getattr(selected_plugin, plugin_property)
			self.data = self.contributions['Data']

		sorted_keys = sorted(self.data, key = self.data.get)
		sorted_contributions = {}

		for key in sorted_keys:
			sorted_contributions[key] = self.data[key]
		
		cmap = plt.get_cmap('plasma')

		label_offset = self.contributions['Total'] / label_offset

		for counter, key in enumerate(sorted_contributions):
			if '-' in key:
				name = key.split('-')[1].strip(' ')
			else:
				name = key

			value = sorted_contributions[key]

			color_value = counter / len(sorted_contributions)
			ax.barh(name, sorted_contributions[key], 
				    color = cmap(color_value), zorder = 5)
			ax.annotate(millify(value), xy = (max(value, 0) + label_offset, counter), 
						va = 'center', ha = 'center')

		ax.barh(make_bold(self.contributions['Table Group']), 
						  self.contributions['Total'], 
						  color = 'darkgreen', zorder = 5)
		ax.annotate(millify(self.contributions['Total']), 
							xy = (self.contributions['Total'] + label_offset, 
								  len(sorted_contributions)), 
							va = 'center', ha = 'center')

		if self.contributions['Table Group'] == 'Total cost of hydrogen':
			x_label_string = x_label_string_H2

		ax.set_xlabel(x_label_string)

		xlim = np.asarray(ax.get_xlim())
		xlim[1] += 2 * label_offset
		ax.set_xlim(xlim)
	
		ax.ticklabel_format(axis = 'x', style = 'sci', scilimits = (0,0), useMathText = True)
		ax.xaxis._update_offset_text_position = types.MethodType(bottom_offset, ax.xaxis)

		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)

		if figure_lean is True:
			figure.execute()
			return figure.fig

